from typing import Any, Optional, List, TypeVar, Generic

from jenesis.config.errors import ConfigurationError

T = TypeVar('T')


def _extract_value(data: Any, path: str) -> Optional[Any]:
    tokens = path.split('.')

    sub_path = tokens[:-1]
    name = tokens[-1]

    for token in sub_path:
        if not isinstance(data, dict):
            return None

        item = data.get(token)
        if item is None:
            return None

        data = item

    if not isinstance(data, dict):
        return None

    return data.get(name)


def _extract_type(data: Any, path: str, obj_type: Generic[T]) -> Optional[T]:
    item = _extract_value(data, path)
    if item is None or (not isinstance(item, obj_type)):
        return None

    return obj_type(item)


def _extract_list_type(data: Any, path: str, obj_type: Generic[T]) -> Optional[List[T]]:
    items = _extract_type(data, path, list)
    return [obj_type(item) for item in items]


def _required(value: Optional[T]) -> T:
    if value is None:
        raise ConfigurationError('unable to extract configuration value')
    return value


def extract_req_int(data: Any, path: str) -> int:
    return _required(_extract_type(data, path, int))


def extract_opt_int(data: Any, path: str) -> Optional[int]:
    return _extract_type(data, path, int)


def extract_req_float(data: Any, path: str) -> float:
    return _required(_extract_type(data, path, float))


def extract_opt_float(data: Any, path: str) -> Optional[float]:
    return _extract_type(data, path, float)


def extract_req_str(data: Any, path: str) -> str:
    return _required(_extract_type(data, path, str))


def extract_opt_str(data: Any, path: str) -> Optional[str]:
    return _extract_type(data, path, str)


def extract_req_dict(data: Any, path: str) -> dict:
    return _required(_extract_type(data, path, dict))


def extract_opt_dict(data: Any, path: str) -> Optional[dict]:
    return _extract_type(data, path, dict)


def extract_req_list(data: Any, path: str) -> List[Any]:
    return _required(_extract_type(data, path, list))


def extract_opt_list(data: Any, path: str) -> Optional[list]:
    return _extract_type(data, path, list)


def extract_req_str_list(data: Any, path: str) -> List[str]:
    return _required(_extract_list_type(data, path, str))



import fnmatch
import os

from typing import Any, List, Optional


def chunks(values: List[Any], batch_size: Optional[int]):
    if batch_size is None:
        yield values
    else:
        for i in range(0, len(values), batch_size):
            yield values[i:i + batch_size]


def get_files(paths: List[str], suffix: str):
    for path in paths:
        for root, _, files in os.walk(path):
            for filename in fnmatch.filter(files, f'*.{suffix}'):
                yield os.path.join(root, filename)


def get_last_modified_timestamp(paths: List[str], suffix: str) -> float:
    return max(
        map(
            os.path.getmtime,
            get_files(paths, suffix)
        ),
        default=0
    )



import sys
import time
from typing import List, Optional, Tuple, Dict

from blessings import Terminal

from jenesis.tasks import Task


class TaskStatusDisplay:
    COMPLETE = -1
    FAILED = -2
    COMPLETE_GLYPH = '✔'
    FAILED_GLYPH = '✖'
    # IN_PROGRESS_GLYPHS = (
    #     '🕛', '🕧', '🕐', '🕜', '🕑', '🕝', '🕒', '🕞',
    #     '🕓', '🕟', '🕔', '🕠', '🕕', '🕡', '🕖', '🕢',
    #     '🕗', '🕣', '🕘', '🕤', '🕙', '🕥', '🕚', '🕦',
    # )

    IN_PROGRESS_GLYPHS = (
        '⠋', '⠙', '⠹', '⠸', '⠼', '⠴', '⠦', '⠧', '⠇', '⠏',
    )

    def __init__(self):
        self._term = Terminal()
        self._first_render = True
        self._name_length = 0
        self._task_progress = {}  # type: Dict[str, int]
        self._task_status_text = {}  # type: Dict[str, Optional[str]]
        self._log = {}

    def update(self, task: Task):
        self._name_length = max(self._name_length, len(task.name))

        # update the progress for this task
        progress = self._task_progress.get(task.name, 0)
        status_text = None
        if task.is_complete:
            progress = self.COMPLETE
        elif task.is_failed:
            progress = self.FAILED
        else:
            progress = (progress + 1) % len(self.IN_PROGRESS_GLYPHS)
            status_text = task.status_text

        # update the status dictionaries
        self._task_progress[task.name] = progress
        self._task_status_text[task.name] = status_text
        self._log[task.name] = task.logs_text

    def render(self):

        # add the terminal blanking on the first render
        if self._first_render:
            for _ in range(len(self._task_progress) + 1):
                print()

        # move the cursor up
        for _ in range(len(self._task_progress)):
            sys.stdout.write(self._term.move_up)

        for name, progress in self._task_progress.items():

            # select the progress text
            if progress == self.COMPLETE:
                glyph = self._term.green(self.COMPLETE_GLYPH)
                progress_text = self._term.green('complete')
            elif progress == self.FAILED:
                glyph = self._term.red(self.FAILED_GLYPH)
                progress_text = self._term.red('FAILED')
            else:
                glyph = self._term.magenta(self.IN_PROGRESS_GLYPHS[progress])
                if self._log[name]:
                    progress_text = self._log[name].splitlines()[-1]
                else:
                    progress_text = self._task_status_text[name]

            # render the status
            width = self._term.width if self._term.is_a_tty else 80
            print(f'  {glyph} {self._term.blue(name)}: {progress_text}'.ljust(width))

        self._first_render = False

    def show_logs(self, task: Task):
        if self._log[task.name]:
            print(f'\n{self._term.green("Logs for")} {self._term.blue(task.name)}:')
            print(self._term.yellow(self._log[task.name]))


def run_tasks(tasks: List[Task], poll_interval: Optional[float] = None) -> Tuple[List[Task], List[Task]]:

    if len(tasks) == 0:
        return [], []

    assert len(tasks) > 0

    poll_interval = poll_interval or 0.1

    display = TaskStatusDisplay()

    completed_tasks = []
    failed_tasks = []

    try:
        while True:

            in_progress_tasks = []
            for task in tasks:

                # query / process the task
                task.poll()

                # update the task
                display.update(task)

                # check the status of the task
                if task.is_complete:
                    completed_tasks.append(task)
                elif task.is_failed:
                    failed_tasks.append(task)
                else:
                    in_progress_tasks.append(task)

            # update the display
            display.render()

            # update the task list
            tasks = in_progress_tasks

            # exit if all the tasks are now complete
            if len(tasks) == 0:
                break

            time.sleep(poll_interval)

        for task in failed_tasks + completed_tasks:
            display.show_logs(task)

    except KeyboardInterrupt:
        print('\nKeyboardInterrupt: shutting down all tasks...')
        for task in tasks:
            task.teardown()
        print('KeyboardInterrupt: shutting down all tasks...complete')
        sys.exit(1)

    return completed_tasks, failed_tasks



from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, List, Optional

class TaskStatus(Enum):
    IDLE = 0
    IN_PROGRESS = 1
    COMPLETE = 2
    FAILED = 3


class Task(ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def status(self) -> TaskStatus:
        pass

    @property
    @abstractmethod
    def logs_text(self) -> str:
        pass

    @property
    @abstractmethod
    def status_text(self) -> str:
        pass

    @property
    def is_idle(self) -> bool:
        return self.status == TaskStatus.IDLE

    @property
    def is_in_progress(self) -> bool:
        return self.status == TaskStatus.IN_PROGRESS

    @property
    def is_complete(self) -> bool:
        return self.status == TaskStatus.COMPLETE

    @property
    def is_failed(self) -> bool:
        return self.status == TaskStatus.FAILED

    @property
    def is_done(self) -> bool:
        return self.is_complete or self.is_failed

    @abstractmethod
    def poll(self):
        pass

    @abstractmethod
    def teardown(self):
        pass



import argparse
import os

from jenesis.config import Config


def run(args: argparse.Namespace):

    project_path = os.getcwd()

    # check that we are actually running the command from the project root
    if not os.path.exists(os.path.join(project_path, "jenesis.toml")):
        # pylint: disable=all
        print("Please run command from project root")
        return

    Config.add_profile(args.profile, args.network)



def add_profile_command(parser):
    profile_cmd = parser.add_parser("profile")
    profile_cmd.add_argument("profile", default="testing", help="The profile to create")
    profile_cmd.add_argument(
        "-n", "--network", default="fetchai-testnet", help="Network to use"
    )
    profile_cmd.set_defaults(handler=run)



import argparse
import os

from jenesis.config import Config


def add_contract_command(parser):

    add_contract_cmd = parser.add_parser("contract")
    add_contract_cmd.add_argument("template", help="The name of the template to use")
    add_contract_cmd.add_argument("name", help="The name of contract to create")
    add_contract_cmd.add_argument('-d','--deployments', nargs='+', help='contract deployments')
    add_contract_cmd.add_argument(
        "-b", "--branch", help="The name of the branch that should be used"
    )
    add_contract_cmd.set_defaults(handler=run_add_contract)


def run_add_contract(args: argparse.Namespace):

    template = args.template
    name = args.name
    branch = args.branch

    project_root = os.path.abspath(os.getcwd())
    contract_root = os.path.join(project_root, "contracts", name)

    cfg = Config.load(project_root)

    # check that we are actually running the command from the project root
    if not os.path.exists(os.path.join(project_root, "jenesis.toml")):
        print("Please run command from project root")
        return False

    # check to see if the contract already exists
    if os.path.exists(contract_root):
        print(f'Contract "{name}" already exists')
        return False

    selected_contract = Config.add_contract(project_root, template, name, branch)

    if selected_contract is None:
        return False

    deployments = args.deployments or [args.name]

    for (profile_name, profile) in cfg.profiles.items():
        network_name = profile.network.name
        for deployment in deployments:
            Config.update_project(os.getcwd(), profile_name, network_name, selected_contract, deployment)

    return True



import argparse

from jenesis.cmd.add.contract import run_add_contract
from jenesis.cmd.add.contract import add_contract_command
from jenesis.cmd.add.profile import add_profile_command
from jenesis.cmd.add.deployment import add_deployment_command


def add_add_command(parser):

    add_cmd = parser.add_parser('add')
    subparsers = add_cmd.add_subparsers()

    add_contract_command(subparsers)
    add_profile_command(subparsers)
    add_deployment_command(subparsers)



import json
from base64 import b64decode
from io import BytesIO
from typing import List, Union

from keyring.backends.SecretService import Keyring as BaseKeyring

from jenesis.keyring.amino.codec import (
    LocalInfo,
    OfflineInfo,
    unmarshal_binary_length_prefixed,
)
from jenesis.keyring.errors import KeychainError


class Keyring(BaseKeyring):
    preferred_collection = "/org/freedesktop/secrets/collection/fetch"


class LinuxKeychainError(KeychainError):
    pass


def query_keychain_item(name: str, keyring_class=Keyring) -> Union[LocalInfo, OfflineInfo]:
    keyring = keyring_class()
    collection = keyring.get_preferred_collection()
    items = list(collection.search_items({"profile": f"{name}.info"}))
    if not items:
        raise LinuxKeychainError("Key not found")

    item = items[0]
    try:
        data = json.loads(item.get_secret())
        binary_data = b64decode(data["Data"])
    except (json.JSONDecodeError, KeyError, TypeError, ValueError) as exc:
        raise LinuxKeychainError("Unable to parse stored entry") from exc

    info = unmarshal_binary_length_prefixed(BytesIO(binary_data))
    if info is None:
        raise LinuxKeychainError("Unable to parse stored entry")
    return info


def query_keychain_items(keyring_class=Keyring) -> List[str]:
    keyring = keyring_class()
    collection = keyring.get_preferred_collection()
    items = list(collection.get_all_items())
    key_names = []
    for item in items:
        account = item.get_label()
        if account.endswith(".info"):
            key_names.append(account[:-5])

    return list(set(key_names))



import argparse
import os

from jenesis.config import Config
from jenesis.contracts.deploy import deploy_contracts
from jenesis.network import network_context


def run(args: argparse.Namespace):

    project_path = os.getcwd()

    # check that we are actually running the command from the project root
    if not os.path.exists(os.path.join(project_path, "jenesis.toml")):
        print("Please run command from project root")
        return 1

    cfg = Config.load(project_path)
    profile = cfg.get_profile(args.profile)
    if profile is None:
        return 1

    with network_context(profile.network, cfg.project_name, profile.name):
        deploy_contracts(cfg, project_path, args.key, profile_name=args.profile)

    return 0


def add_deploy_command(parser):
    deploy_cmd = parser.add_parser("deploy")
    deploy_cmd.add_argument(
        "-p", "--profile", default=None, help="The profile to deploy"
    )
    deploy_cmd.add_argument("key", nargs="?", help="Deployer Key for all contracts")
    deploy_cmd.set_defaults(handler=run)



import argparse
import os
import sys

from jenesis.config import Config, Profile
from jenesis.network import network_context

from .shell import get_profile, load_shell_globals


PROJECT_PATH = os.getcwd()


def run(args: argparse.Namespace):
    cfg = Config.load(PROJECT_PATH)
    profile: Profile = get_profile(cfg, args)

    with network_context(profile.network, cfg.project_name, profile.name):
        shell_globals = load_shell_globals(cfg, profile)

        # Make the script args available inside the script
        sys.argv = [args.script_path] + args.args
        shell_globals['sys'] = sys

        with open(args.script_path, encoding="utf-8") as file:
            code = compile(
                file.read(),
                os.path.basename(args.script_path),
                'exec',
            )
            exec(code, shell_globals)


def add_run_command(parser):
    run_cmd = parser.add_parser("run")
    run_cmd.add_argument(
        "-p", "--profile", default=None, help="The profile to use"
    )
    run_cmd.add_argument("script_path", help="The path to the script to run")
    run_cmd.add_argument(
        "args", nargs= "*", help="The args to pass to the script"
    )
    run_cmd.set_defaults(handler=run)



import hashlib
import json
import os
import shutil
import struct
import subprocess
from dataclasses import dataclass
from tempfile import mkdtemp
from typing import Any, Dict, List, Optional

import toml
from cosmpy.crypto.address import Address
from jenesis.config.errors import ConfigurationError
from jenesis.config.extract import (extract_opt_dict, extract_opt_int,
                                    extract_opt_str, extract_req_dict,
                                    extract_req_str, extract_req_str_list,
                                    extract_opt_list)
from jenesis.contracts import Contract
from jenesis.contracts.detect import detect_contracts, parse_contract
from jenesis.network import (Network, fetchai_localnode_config,
                             fetchai_testnet_config, fetchai_mainnet_config)


TEMPLATE_GIT_URL = "https://github.com/fetchai/jenesis-templates.git"
DEFAULT_KEYRING_BACKEND = "os"


@dataclass
class Deployment:
    name: str # internal: the name for the deployment
    contract: str  # internal: the name of the contract to deploy
    network: str  # internal: the name of the network to deploy to
    deployer_key: str  # config: the name of the key to use for deployment
    init: Any  # config: init parameters for the contract
    init_funds: Optional[str] # config: funds to be sent with instantiation msg
    checksum: Optional[str]  # lock: the checksum digest to detect configuration changes
    digest: Optional[str]  # lock: the contract of the deployed contract
    address: Optional[Address]  # lock: the address of the deployed contract
    code_id: Optional[int]  # lock: the code of the deployed contract

    def reset_metadata(self):
        self.checksum = None
        self.digest = None
        self.address = None
        self.code_id = None

    def compute_checksum(self) -> str:
        hasher = hashlib.sha256()

        def update(value: Optional[Any]):
            encoded_value = "" if value is None else str(value)

            hasher.update(struct.pack(">Q", len(encoded_value)))
            hasher.update(encoded_value.encode())

        canonical_init_args = json.dumps(self.init, sort_keys=True)

        update(self.network)
        update(self.deployer_key)
        update(canonical_init_args)
        update(self.digest)
        update(self.address)
        update(self.code_id)

        return hasher.hexdigest()

    def __repr__(self) -> str:
        return f'{self.contract}: {self.address}'

    def is_configuration_out_of_date(self) -> bool:
        return self.checksum != self.compute_checksum()

    def to_lockfile(self) -> Any:
        data = {
            "checksum": self.compute_checksum(),
        }

        if self.digest is not None:
            data["digest"] = str(self.digest)
        if self.address is not None:
            data["address"] = str(self.address)
        if self.code_id is not None:
            data["code_id"] = self.code_id

        return data


@dataclass
class Profile:
    name: str
    network: Network
    deployments: Dict[str, Deployment]
    default: bool = False

    def to_lockfile(self) -> Any:
        return {
            name: deployment.to_lockfile()
            for name, deployment in self.deployments.items()
        }


@dataclass
class Config:
    project_name: str
    project_authors: List[str]
    profiles: Dict[str, Profile]
    keyring_backend: str = "os"

    def update_deployment(
        self,
        profile_name: str,
        deployment_name: str,
        digest: str,
        code_id: int,
        address: Address,
    ):
        profile = self.profiles.get(profile_name)
        if profile is None:
            raise ConfigurationError(f"unable to lookup profile {profile_name}")

        deployment = profile.deployments.get(deployment_name)
        assert deployment is not None, f"Deployment not found: {deployment_name}"

        # update the contract if necessary
        if digest is not None:
            deployment.digest = str(digest)
        if code_id is not None:
            deployment.code_id = int(code_id)
        if address is not None:
            deployment.address = Address(address)

        profile.deployments[deployment_name] = deployment
        self.profiles[profile_name] = profile

    def get_default_profile(self) -> str:
        for (name, profile) in self.profiles.items():
            if profile.default:
                return name
        return self.profiles.keys()[0]

    def get_profile(self, profile_name: Optional[str]) -> Optional[Profile]:
        if profile_name is None:
            profile_name = self.get_default_profile()
        else:
            if profile_name not in self.profiles:
                print(f'Invalid profile name. Expected one of {",".join(self.profiles.keys())}')
                return None
        return self.profiles[profile_name]

    @classmethod
    def load(cls, path: str) -> "Config":
        project_file_path = os.path.join(path, "jenesis.toml")
        lock_file_path = os.path.join(path, "jenesis.lock")

        if not os.path.isfile(project_file_path):
            raise ConfigurationError('Missing project file: "jenesis.toml"')
        project_contents = toml.load(project_file_path)

        if os.path.isfile(lock_file_path):
            lock_file_contents = toml.load(lock_file_path)
        else:
            lock_file_contents = {}

        return cls._loads(project_contents, lock_file_contents)

    @classmethod
    def _loads(cls, project_contents: Any, lock_file_contents: Any) -> "Config":

        # extract the profiles section
        config_profiles = project_contents.get("profile", {})
        if not isinstance(config_profiles, dict):
            raise ConfigurationError("invalid profile configuration section")

        lock_profiles = lock_file_contents.get("profile", {})
        if not isinstance(lock_profiles, dict):
            raise ConfigurationError("invalid profile lock section")

        profiles = {}
        for name, config_item in config_profiles.items():
            lock_item = lock_profiles.get(name, {})

            profile = cls._parse_profile(name, config_item, lock_item)
            profiles[profile.name] = profile

        keyring_backend = extract_opt_str(
            project_contents, "project.keyring_backend"
        ) or DEFAULT_KEYRING_BACKEND

        return Config(
            project_name=extract_req_str(project_contents, "project.name"),
            project_authors=extract_req_str_list(project_contents, "project.authors"),
            profiles=profiles,
            keyring_backend=keyring_backend,
        )

    @classmethod
    def _parse_profile(cls, name: str, profile: Any, lock_profile: Any) -> Profile:
        if not isinstance(profile, dict):
            raise ConfigurationError(
                "profile configuration invalid, expected dictionary"
            )
        if not isinstance(lock_profile, dict):
            raise ConfigurationError(
                "profile lock configuration invalid, expected dictionary"
            )

        network = Network(**extract_req_dict(profile, "network"))

        profile_contracts = profile.get("contracts", {})
        if not isinstance(profile_contracts, dict):
            raise ConfigurationError("invalid contracts section in config")

        deployments = {}
        if "contracts" in profile:
            for deployment_name, contract_cfg in profile_contracts.items():
                deployment_lock = lock_profile.get(deployment_name, {})

                deployment = cls._parse_contract_config(
                    deployment_name, contract_cfg, network.name, deployment_lock
                )
                deployments[deployment_name] = deployment

        is_default = False
        if "default" in profile:
            if profile["default"]:
                is_default = True

        return Profile(
            name=str(name),
            network=network,
            deployments=deployments,
            default=is_default,
        )

    @classmethod
    def _parse_contract_config(
        cls, deployment_name: str, contract_cfg: dict, network: str, lock: Any
    ) -> Deployment:
        if not isinstance(contract_cfg, dict):
            raise ConfigurationError(
                "contract configuration invalid, expected dictionary"
            )
        if not isinstance(lock, dict):
            raise ConfigurationError(
                "contract lock configuration invalid, expected dictionary"
            )

        def opt_address(value: Optional[str]) -> Optional[Address]:
            return None if value is None else Address(value)

        return Deployment(
            name=str(deployment_name),
            contract=extract_req_str(contract_cfg, "contract"),
            network=str(network),
            init=extract_opt_dict(contract_cfg, "init"),
            deployer_key=extract_req_str(contract_cfg, "deployer_key"),
            init_funds=extract_opt_str(contract_cfg, "init_funds"),
            digest=extract_opt_str(lock, "digest"),
            address=opt_address(extract_opt_str(lock, "address")),
            code_id=extract_opt_int(lock, "code_id"),
            checksum=extract_opt_str(lock, "checksum"),
        )

    def save(self, path: str):
        contents = {
            "profile": {
                name: profile.to_lockfile() for name, profile in self.profiles.items()
            }
        }

        lock_file_path = os.path.join(path, "jenesis.lock")
        with open(lock_file_path, "w", encoding="utf-8") as lock_file:
            toml.dump(contents, lock_file)

    @staticmethod
    def create_project(path: str, profile: str, network_name: str):
        user_name = subprocess.getoutput("git config user.name")
        user_email = subprocess.getoutput("git config user.email")
        authors = [f"{user_name} <{user_email}>"]

        # take the project name directly from the base name of the project
        project_root = os.path.abspath(path)
        project_name = os.path.basename(project_root)

        # detect contract source code and add placeholders for key contract data
        contracts = detect_contracts(project_root) or []

        deployments = {contract.name: Deployment(contract.name, contract.name,
            network_name, "", {arg: "" for arg in contract.init_args()},
            "", None, None, None, None,
        ) for contract in contracts}

        if network_name == "fetchai-testnet":
            net_config = fetchai_testnet_config()
        elif network_name == "fetchai-localnode":
            net_config = fetchai_localnode_config()
        elif network_name == "fetchai-mainnet":
            net_config = fetchai_mainnet_config()
        else:
            raise ConfigurationError("Network name not recognized")

        network = {"name": network_name}
        network.update(vars(net_config))

        profiles = {
            profile: {
                "network": network,
                "contracts": {name: vars(cfg) for (name, cfg) in deployments.items()},
                "default": True,
            }
        }

        data = {
            "project": {"name": project_name, "authors": authors, "keyring_backend": "os"},
            "profile": profiles,
        }

        project_configuration_file = os.path.join(project_root, "jenesis.toml")
        project_git_keep_files = [
            os.path.join(project_root, "contracts", ".gitkeep"),
        ]

        # create the configuration file
        os.makedirs(os.path.dirname(project_configuration_file), exist_ok=True)
        with open(project_configuration_file, "w", encoding="utf-8") as toml_file:
            toml.dump(data, toml_file)

        # create all the git kep files
        for git_keep_path in project_git_keep_files:
            os.makedirs(os.path.dirname(git_keep_path), exist_ok=True)

            with open(git_keep_path, "a", encoding="utf-8"):
                pass

    @staticmethod
    def update_project(path: str, profile: str, network_name: str, contract: Contract, deployment_name: str):

        # take the project name directly from the base name of the project
        project_root = os.path.abspath(path)

        deployment = Deployment(deployment_name, contract.name,
            network_name, "", {arg: "" for arg in contract.init_args()},
            "", None, None, None, None)

        data = toml.load("jenesis.toml")

        data["profile"][profile]["contracts"][deployment_name] = vars(deployment)
        project_configuration_file = os.path.join(project_root, "jenesis.toml")

        with open(project_configuration_file, "w", encoding="utf-8") as toml_file:
            toml.dump(data, toml_file)

    @staticmethod
    def update_key(path: str, profile: str, deployment_name: str, key: str):

        # take the project name directly from the base name of the project
        project_root = os.path.abspath(path)

        input_file_name = "jenesis.toml"
        path = os.path.join(os.getcwd(), input_file_name)
        with open(path, encoding="utf-8") as toml_file:
            data = toml.load(toml_file)

        data["profile"][profile]["contracts"][deployment_name]["deployer_key"] = key
        project_configuration_file = os.path.join(project_root, "jenesis.toml")

        with open(project_configuration_file, "w", encoding="utf-8") as toml_file:
            toml.dump(data, toml_file)

    @staticmethod
    def add_profile(profile: str, network_name: str):

        data = toml.load("jenesis.toml")

        if network_name == "fetchai-testnet":
            net_config = fetchai_testnet_config()
        elif network_name == "fetchai-localnode":
            net_config = fetchai_localnode_config()
        elif network_name == "fetchai-mainnet":
            net_config = fetchai_mainnet_config()
        else:
            raise ConfigurationError("Network name not recognized")

        network = {"name": ""}
        network.update(vars(net_config))

        default_profile = list(data["profile"].keys())[0]

        contract_data = data["profile"][default_profile]["contracts"]

        for deployment_name in contract_data.keys():
            contract_data[deployment_name]["network"] = network_name

        data["profile"][profile] = {
            "network": network,
            "contracts": contract_data,
        }

        output_file_name = "jenesis.toml"
        with open(output_file_name, "w") as toml_file:
            # pylint: disable=all
            toml.dump(data, toml_file)

    @staticmethod
    def add_contract(project_root: str, template: str, name: str, branch: str) -> Contract:

        contract_path = os.path.join(project_root, 'contracts', name)

        # create the temporary clone folder
        temp_clone_path = mkdtemp(prefix="jenesis-", suffix="-tmpl")

        # clone the templates folder out in the temporary file
        print("Downloading template...")
        cmd = ["git", "clone", "--single-branch"]
        if branch is not None:
            cmd += ["--branch", branch]
        cmd += [TEMPLATE_GIT_URL, "."]
        with open(os.devnull, "w", encoding="utf8") as null_file:
            subprocess.check_call(
                cmd, stdout=null_file, stderr=subprocess.STDOUT, cwd=temp_clone_path
            )

        # find the target contract
        available_templates = os.listdir(os.path.join(temp_clone_path, "contracts"))
        contract_template_path = os.path.join(temp_clone_path, "contracts", template)
        if not os.path.isdir(contract_template_path):
            print(f"Unknown template {template}: expecting one of {set(available_templates)}")
            return
        print("Downloading template...complete")

        # process all the files as part of the template
        print("Rendering template...")
        for root, _, files in os.walk(contract_template_path):
            for filename in files:
                file_path = os.path.join(root, filename)
                rel_path = os.path.relpath(file_path, contract_template_path)

                output_filepath = os.path.join(contract_path, rel_path)
                os.makedirs(os.path.dirname(output_filepath), exist_ok=True)
                with open(file_path, "r", encoding="utf8") as input_file:
                    with open(output_filepath, "w", encoding="utf8") as output_file:
                        contents = input_file.read()

                        # replace the templating parameters here
                        if file_path.endswith(".toml"):
                            contents = contents.replace("<<NAME>>", name)
                        else:
                            contents = contents.replace("<<NAME>>", name.replace("-", "_"))

                        output_file.write(contents)
        print("Rendering template...complete")

        # clean up the temporary folder
        shutil.rmtree(temp_clone_path)

        return parse_contract(project_root, name)



from pathlib import Path
from typing import Dict, List, Union

from jwskate import JweCompact

from jenesis.keyring.amino.codec import LocalInfo, OfflineInfo
from jenesis.keyring.errors import KeychainError
from jenesis.keyring.linux import (
    query_keychain_item as linux_query_keychain_item,
)
from jenesis.keyring.linux import (
    query_keychain_items as linux_query_keychain_items,
)


class KeychainItem:
    DEFAULT_COSMOS_TEST_KEYRING_PASSWORD = "test"

    def __init__(self, filepath: Path):
        self._filepath = Path(filepath)

    def get_label(self) -> str:
        return self._filepath.name

    def get_secret(self):
        return JweCompact(self._filepath.read_bytes()).decrypt_with_password(
            self.DEFAULT_COSMOS_TEST_KEYRING_PASSWORD
        )


class Collection:
    def __init__(self, path: Path):
        self._path = path

    def get_all_items(self) -> List[KeychainItem]:
        return [KeychainItem(i) for i in self._path.iterdir() if i.is_file()]

    def search_items(self, query: Dict[str, str]) -> List[KeychainItem]:
        if "profile" not in query:
            raise TestKeychainError("only `profile` query supported")
        return [
            i
            for i in self.get_all_items()
            if i.get_label() == query["profile"]
        ]


class KeyringTestBackend:  # pylint: disable=too-few-public-methods
    dir = ".fetchd/keyring-test/"

    def get_preferred_collection(self) -> Collection:
        return Collection(Path.home() / self.dir)


class TestKeychainError(KeychainError):
    pass


def query_keychain_item(name: str) -> Union[LocalInfo, OfflineInfo]:
    return linux_query_keychain_item(
        name=name, keyring_class=KeyringTestBackend
    )


def query_keychain_items() -> List[str]:
    return linux_query_keychain_items(keyring_class=KeyringTestBackend)



import json
import os

from typing import List, Optional

from jenesis.contracts import Contract
from jenesis.contracts.build import ContractBuildTask
from jenesis.tasks.utils import chunks, get_last_modified_timestamp
from jenesis.tasks.monitor import run_tasks

SCHEMA_BUILD_STEPS = ["cargo schema"]


class ContractSchemaTask(ContractBuildTask):

    def __init__(self, contract: Contract, rebuild: bool):
        super().__init__(contract, False, rebuild, False)
        self._build_steps = SCHEMA_BUILD_STEPS

        # set working directory according to whether project is a cargo workspace
        if contract.source_path == contract.cargo_root:
            self._working_dir = '/code'
        else: # cargo workspace case
            self._working_dir = os.path.join(
                '/code',
                'contracts',
                os.path.basename(contract.source_path)
            )

        self._in_progress_text = "Generating schemas..."

    def _is_out_of_date(self) -> bool:
        #  pylint: disable=duplicate-code
        if self._rebuild:
            return True

        # determine the timestamp of the compiled contract
        if os.path.isfile(self.contract.binary_path):
            compiled_contract_timestamp = os.path.getmtime(self.contract.binary_path)
        else:
            compiled_contract_timestamp = 0

        # determine the timestamp of the contract schema files
        schema_path = os.path.join(self.contract.source_path, 'schema')
        contract_schema_timestamp = get_last_modified_timestamp([schema_path], 'json')

        return compiled_contract_timestamp > contract_schema_timestamp


def generate_schemas(
    contracts: List[Contract],
    batch_size: Optional[int] = None,
    rebuild: Optional[bool] = False,
):
    """
    Will attempt to build all the specified contract schemas (provided they are out of date)

    :param contracts: The list of contracts to build schemas for
    :param batch_size: The max number of builds to do in parallel. If None then will attempt to all in parallel
    :param rebuild: Whether to force a rebuild of the contract schemas
    :return:
    """
    # create all the tasks to be done
    tasks = list(
        map(
            ContractSchemaTask,
            contracts,
            [rebuild] * len(contracts),
        )
    )

    # run the tasks (in batches if configured)
    for batch in chunks(tasks, batch_size=batch_size):
        run_tasks(batch)

    for contract in contracts:
        contract.schema = load_contract_schema(contract.source_path)
        contract.update_schema()


def load_contract_schema(source_path: str) -> dict:
    schema_folder = os.path.join(source_path, 'schema')
    if not os.path.isdir(schema_folder):
        return {}

    schema = {}
    for filename in os.listdir(schema_folder):
        if filename.endswith('.json'):
            msg_name = os.path.splitext(os.path.basename(filename))[0]
            full_path = os.path.join(schema_folder, filename)
            with open(full_path, 'r', encoding="utf-8") as msg_schema_file:
                msg_schema = json.load(msg_schema_file)
            schema[msg_name] = msg_schema
    return schema



import graphlib as gl
import os
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import TimeoutError as ConcurrentTimeoutError
from typing import Dict, List, Optional, Set, Union

from cosmpy.aerial.client import LedgerClient
from cosmpy.aerial.contract import LedgerContract
from cosmpy.aerial.wallet import LocalWallet, Wallet
from cosmpy.crypto.address import Address
from cosmpy.crypto.keypairs import PrivateKey
from jenesis.config import Config, Deployment, Profile
from jenesis.contracts import Contract
from jenesis.contracts.detect import detect_contracts
from jenesis.contracts.monkey import MonkeyContract
from jenesis.keyring import (LocalInfo, query_keychain_item,
                             query_keychain_items)
from jenesis.tasks import Task, TaskStatus
from jenesis.tasks.monitor import run_tasks


# Recursive function to insert deployed contract address into instantiation msg
def insert(data: Union[Dict, List], contract_name: str, address: str):
    for key, value in data.items() if isinstance(data, dict) else enumerate(data):
        if value == ("$" + contract_name):
            data[key] = address
        elif isinstance(value, (dict, list)):
            insert(value, contract_name, address)


def insert_address(contract_address_names: List[str], deployment: Deployment, profile: Profile) -> dict:

    deployment_names = list(profile.deployments.keys())
    init_data = deployment.init

    # iterate over the addresses to insert
    for name in contract_address_names:
        if name in deployment_names:
            deployed_contract_address = profile.deployments[name].address

            assert deployed_contract_address is not None, f"Contract {name} address not found"

            # insert deployed contract address in instantiation msg
            insert(init_data, name, str(deployed_contract_address))

    return init_data

def retreive_init_addresses(data: Union[Dict, List], contract_names: List[str], init_addresses: List[str]):
    for _, value in data.items() if isinstance(data, dict) else enumerate(data):
        if isinstance(value, str) and len(value) > 0:
            if value[0] == "$" and value[1:] in contract_names:
                init_addresses.append(value[1:])
                contract_names.remove(value[1:])
        elif isinstance(value, (dict, list)):
            retreive_init_addresses(value, contract_names,init_addresses)
    return init_addresses

def load_keys(key_names: Set[str], cfg: Config) -> Dict[str, PrivateKey]:
    keys = {}
    available_key_names = set(query_keychain_items(cfg.keyring_backend))
    for key_name in key_names:

        if key_name not in available_key_names:
            print(f"Key not found: {key_name}")
            continue

        info = query_keychain_item(key_name, cfg.keyring_backend)
        if not isinstance(info, LocalInfo):
            print(f"Failed to retrieve local key: {key_name}")
            continue

        keys[key_name] = PrivateKey(info.private_key)
    return keys


class DeployContractTask(Task):
    def __init__(self, project_path: str, cfg: Config, profile: Profile, contract: Contract,
                 deployment: Deployment, client: LedgerClient, wallet: Wallet):
        self._project_path = project_path
        self._cfg = cfg
        self._profile = profile
        self._contract = contract
        self._deployment = deployment
        self._client = client
        self._wallet = wallet

        # task status
        self._status = TaskStatus.IDLE
        self._status_text = ''
        self._logs = ''

        # background worker
        self._executor = ThreadPoolExecutor(max_workers=1)

        # state machine state
        self._state = 'idle'
        self.ledger_contract = None  # type: Optional[LedgerContract]
        self.contract_address = None  # type: Optional[Address]
        self._future = None

    @property
    def name(self) -> str:
        return self._deployment.name

    @property
    def status(self) -> TaskStatus:
        return self._status

    @property
    def status_text(self) -> str:
        return self._status_text

    @property
    def logs_text(self) -> str:
        return self._logs

    def poll(self):
        if self._state == 'idle':
            self._schedule_build_contract()
        elif self._state == 'wait-for-ledger-contract':
            self._wait_for_ledger_contract()
        elif self._state == 'schedule-deployment':
            self._schedule_deploy_contract()
        elif self._state == 'wait-for-deployment':
            self._wait_for_contract_deployment()
        elif self._state == 'failed':
            self._failed()
        elif self._state == 'complete':
            self._complete()
        else:
            assert False, "bad state"

    def _schedule_build_contract(self):
        assert self._future is None

        def action():
            return MonkeyContract(
                self._contract,
                self._client,
                code_id=self._deployment.code_id,
                address=self._deployment.address
            )

        self._future = self._executor.submit(action)
        self._state = 'wait-for-ledger-contract'
        self._status = TaskStatus.IN_PROGRESS
        self._status_text = '(1/2) Determining contract parameters...'

    def _wait_for_ledger_contract(self):
        try:
            self.ledger_contract = self._future.result(timeout=0.05)
            self._state = 'schedule-deployment'
            self._future = None
        except ConcurrentTimeoutError:
            pass

    def _schedule_deploy_contract(self):
        assert self._future is None
        assert self.ledger_contract is not None

        def action():
            return self.ledger_contract._deploy( #  pylint: disable=W0212
                args=self._deployment.init,
                sender=self._wallet,
                admin_address=self._wallet.address(),
                funds=self._deployment.init_funds,
            )

        self._future = self._executor.submit(action)
        self._state = 'wait-for-deployment'
        self._status = TaskStatus.IN_PROGRESS
        self._status_text = '(2/2) Deploying contract...'

    def _wait_for_contract_deployment(self):
        try:
            self.contract_address = self._future.result(timeout=0.05)
            self._state = 'complete'
            self._future = None
        except ConcurrentTimeoutError:
            pass

    def _complete(self):
        # update the configuration and save it to disk
        self._cfg.update_deployment(self._profile.name, self._deployment.name,
                                    self.ledger_contract.digest.hex(),
                                    self.ledger_contract.code_id, self.contract_address)
        self._cfg.save(self._project_path)

        self._finished(True)

    def _failed(self):
        self._finished(False)

    def _finished(self, success: bool):
        self._executor.shutdown(wait=True)
        if success:
            self._status = TaskStatus.COMPLETE
        else:
            self._status = TaskStatus.FAILED
        self._status_text = ''

    def teardown(self):
        pass


def deploy_contracts(cfg: Config, project_path: str, deployer_key: Optional[str], profile_name: Optional[str] = None):
    if profile_name is None:
        profile_name = cfg.get_default_profile()

    profile = cfg.profiles[profile_name]

    project_contracts = {contract.name: contract for contract in detect_contracts(project_path)}
    deployments = profile.deployments

    init_addresses = {}

    # iterate over the addresses to insert
    for (name, deployment) in deployments.items():
        init_data = deployment.init
        deployment_names = list(deployments.keys())

        # retreive init addresses for each contract
        contract_init_addresses = retreive_init_addresses(init_data, deployment_names, [])

        # add all contract init addresses in one dictionary
        init_addresses[name] = set(contract_init_addresses)

    # sort the contracts to be deployed in the right order
    sorter = gl.TopologicalSorter(init_addresses)
    deployment_order = list(sorter.static_order())

    # load all the keys required for this operation
    key_names = {deployment.deployer_key for deployment in deployments.values()} | {deployer_key}
    keys = load_keys(key_names, cfg)

    for deployment_name in deployment_order:
        deployment = deployments[deployment_name]

        # ensure specified contract is in project
        if deployment.contract not in project_contracts:
            print(f"Contract {deployment_name} not found in project")
            continue
        contract = project_contracts[deployment.contract]

        # ensure that contract has been compiled first
        if not os.path.isfile(contract.binary_path):
            print(f"No contract binary found for {contract.name}. Please run 'jenesis compile' first.")
            continue

        if deployer_key is not None:

            if deployer_key not in keys:
                print(f"Skipping {deployment_name}: deployer key {deployer_key} not available")
                continue

            deployment.deployer_key = deployer_key
            Config.update_key(os.getcwd(), profile_name, deployment_name, deployer_key)

        if deployment.address is not None:
            if not deployment.is_configuration_out_of_date():
                print(f"Skipping {deployment_name}: configuration is up to date")
                continue

            if contract.digest() == deployment.digest:
                print(f"Skipping {deployment_name}: digest has not changed")
                continue

        client = LedgerClient(profile.network)

        deployment.address = None  # clear the old address

        contract_address_names = list(init_addresses[deployment_name])

        deployment.init = insert_address(contract_address_names, deployment, profile)

        # lookup the wallet key
        wallet = LocalWallet(keys[deployment.deployer_key])

        # create the deployment task
        task = DeployContractTask(
            project_path,
            cfg,
            profile,
            contract,
            deployment,
            client,
            wallet,
        )

        # run the deployment task
        run_tasks([task])



from cosmpy.crypto.address import Address

from jenesis.config import Config
from jenesis.contracts.monkey import ContractObserver


class DeploymentUpdater(ContractObserver):
    def __init__(self, cfg: Config, project_path: str, profile: str, deployment_name: str):
        self._cfg = cfg
        self._project_path = str(project_path)
        self._profile = str(profile)
        self._deployment_name = str(deployment_name)

    def on_code_id_update(self, code_id: int):
        self._cfg.update_deployment(self._profile, self._deployment_name, None, code_id, None)
        self._cfg.save(self._project_path)

    def on_contract_address_update(self, address: Address):
        self._cfg.update_deployment(self._profile, self._deployment_name, None, None, address)
        self._cfg.save(self._project_path)



import argparse
import os

from jenesis.config import Config


def run(args: argparse.Namespace):

    project_path = os.getcwd()

    # check if the project has been already initialized
    if os.path.exists(os.path.join(project_path, "jenesis.toml")):
        # pylint: disable=all
        print("Project already initialized")
        return

    Config.create_project(project_path, args.profile, args.network)


def add_init_command(parser):
    init_cmd = parser.add_parser("init")
    init_cmd.add_argument(
        "-p", "--profile", default="testing", help="The profile to initialize"
    )
    init_cmd.add_argument(
        "-n", "--network", default="fetchai-testnet", help="Network to use"
    )
    init_cmd.set_defaults(handler=run)


import os
from typing import List, Optional

import toml

from jenesis.contracts import Contract
from jenesis.contracts.schema import load_contract_schema


def is_workspace(path: str) -> bool:
    contracts_folder = os.path.join(path, 'contracts')
    if not os.path.isdir(contracts_folder):
        return False

    expected_files = (
        'Cargo.toml',
        'Cargo.lock',
    )
    for expected_file in expected_files:
        if not os.path.isfile(os.path.join(path, expected_file)):
            return False

    return True


def parse_contract(path: str, name: str) -> Contract:
    contracts_folder = os.path.join(path, 'contracts')
    cargo_file_path = os.path.join(contracts_folder, name, 'Cargo.toml')
    with open(cargo_file_path, 'r', encoding="utf-8") as cargo_file:
        cargo_contents = toml.load(cargo_file)

    # extract the contract name and replace hyphens with underscores
    contract_name = cargo_contents['package']['name'].replace("-", "_")

    source_path = os.path.abspath(os.path.join(contracts_folder, name))
    if is_workspace(path):
        cargo_root = path
    else:
        cargo_root = os.path.join(contracts_folder, name)
    artifacts_path = os.path.join(cargo_root, 'artifacts')
    binary_path = os.path.abspath(os.path.join(artifacts_path, f'{contract_name}.wasm'))

    schema = load_contract_schema(source_path)

    return Contract(
        name=contract_name,
        source_path=source_path,
        binary_path=binary_path,
        cargo_root=cargo_root,
        schema=schema,
    )


def detect_contracts(path: str) -> Optional[List[Contract]]:
    contracts_folder = os.path.join(path, 'contracts')
    if not os.path.isdir(contracts_folder):
        return None

    def is_contract(folder_name: str) -> bool:
        contract_path = os.path.join(contracts_folder, folder_name)
        if not os.path.isdir(contract_path):
            return False

        expected_files = (
            'Cargo.toml',
        )
        for expected_file in expected_files:
            if not os.path.isfile(os.path.join(contract_path, expected_file)):
                return False

        return True

    contract_list = list(filter(
        is_contract,
        os.listdir(contracts_folder)
    ))

    contracts = map(
        parse_contract,
        [path] * len(contract_list),
        contract_list,
    )

    return list(contracts)
